(* $Id: pa_m4.ml 1731 2007-10-29 11:00:12Z roglo $ *)

#load "q_MLast.cmo";;
#load "pa_extend.cmo";;
#load "pa_macro.cmo";;

open Pcaml

let space = parser [< '' ' | '\t' | '\n' as c >] -> c
let letter = parser [< ''a'..'z' | 'A'..'Z' | '_' as c >] -> c
let digit = parser [< ''0'..'9' | '_' as c >] -> c

let rec leading_space c =
  parser
    [< c1 = space; s = leading_space c1 >] -> String.make 1 c ^ s
  | [< >] -> String.make 1 c

let rec name c =
  parser
    [< c1 = letter; s = name c1 >] -> String.make 1 c ^ s
  | [< c1 = digit; s = name c1 >] -> String.make 1 c ^ s
  | [< >] -> String.make 1 c

let rec quoted_string =
  parser
    [< ''\''>]  -> ""
  | [< ''`'; s1 = quoted_string; s2 = quoted_string >]  -> "`" ^ s1 ^ "'" ^ s2
  | [< 'c; s = quoted_string >] -> String.make 1 c ^ s

let rec comment =
  parser
    [< ''\n'>]  -> ""
  | [< 'c; s = comment >] -> String.make 1 c ^ s

let rec other c =
  parser
    [< c1 = space; s = other c1 >] -> String.make 1 c ^ s
  | [< c1 = letter; s = other c1 >] -> String.make 1 c ^ s
  | [< c1 = digit; s = other c1 >] -> String.make 1 c ^ s
  | [< >] -> String.make 1 c

let basic_lexer =
  parser
    [< c = space; s = leading_space c >] -> Some ("LEADSP", s)
  | [< c = letter; s = name c >] -> Some ("NAME", s)
  | [< ''`'; s = quoted_string >] -> Some ("QUOTSTR", String.escaped s)
  | [< ''(' >] -> Some ("", "(")
  | [< '')' >] -> Some ("", ")")
  | [< '',' >] -> Some ("", ",")
  | [< ''#'; s = comment >] -> Some ("COMMENT", s)
  | [< 'c; s = other c >] -> Some ("OTHER", s)
  | [< >] -> Some ("EOI", "")

let rec skip_to_eoln =
  parser
    [< ''\n' >] -> ()
  | [< '_; x = skip_to_eoln >] -> x
  | [< >] -> ()

let lexer_func cstrm =
  let loct = ref [| |] in
  let redim i =
    if i >= Array.length !loct then
      let nloct = Array.create (2 * Array.length !loct + 1) Token.dummy_loc in
      Array.blit !loct 0 nloct 0 (Array.length !loct);
      loct := nloct
    else ()
  in
  let tstrm =
    let lnum = ref 1 in
    let bol = ref 0 in
    let last_was_eol = ref false in
    let cs =
      Stream.from
        (fun i ->
           if !last_was_eol then begin
             last_was_eol := false; incr lnum; bol := i
           end;
           match cstrm with parser
             [< ''\n' >] -> last_was_eol := true; Some '\n'
           | [< 'c >] -> Some c
           | [< >] -> None)
    in
    let make_pos p =
      IFDEF CAMLP5 THEN (!lnum, !bol, p)
      ELSE
        {Lexing.pos_fname = !Pcaml.input_file;
         Lexing.pos_lnum = !lnum;
         Lexing.pos_bol = !bol;
         Lexing.pos_cnum = p}
      END
    in
    let make_loc (bp, ep) =
      IFDEF CAMLP5 THEN
        let (lnum, bol, bp) = bp in
        let (_, _, ep) = ep in
        Stdpp.make_lined_loc lnum bol (bp, ep)
      ELSE (bp, ep)
      END
    in
    let rec lexer () =
      let bp = make_pos (Stream.count cs) in
      match basic_lexer cs with
        Some ("NAME", "dnl") ->
          skip_to_eoln cs;
          lexer ()
      | t ->
          let ep = make_pos (Stream.count cs) in
          (t, make_loc (bp, ep))
    in
    Stream.from
      (fun i ->
         let (t, loc) = lexer () in
         redim i;
         !loct.(i) <- loc;
         t)
  in
  let floc i = !loct.(i) in
  (tstrm, floc)

let using_tok =
  function
    ("LEADSP", _) | ("NAME", _) | ("QUOTSTR", _) | ("COMMENT", _)
  | ("", "(") | ("", ")") | ("", ",") | ("OTHER", _) | ("EOI", "") -> ()
  | p ->
      failwith
        ("The token " ^ Token.lexer_text p ^ " is not recognized by the lexer")

let lexer =
  {Token.tok_func = lexer_func;
   Token.tok_using = using_tok;
   Token.tok_removing = (fun _ -> ());
   Token.tok_match = Token.default_match;
   Token.tok_text = Token.lexer_text;
   Token.tok_comm = None}
in
Grammar.Unsafe.gram_reinit Pcaml.gram lexer;;
(*
Pcaml.inter_phrases := Some "\n"
*)

type m4 =
    Macro of MLast.loc * string * (MLast.loc * (string * string)) list
  | Comment of MLast.loc * string
  | QuotedString of MLast.loc * string
  | LeadingSpace of MLast.loc * string

let encl_loc loc1 loc2 =
  IFDEF CAMLP5 THEN Stdpp.encl_loc loc1 loc2
  ELSE (fst loc1, snd loc2)
  END

let m4tosi =
  function
    Macro (loc, m, []) ->
      (<:str_item< macro $str:m$ >>, loc)
  | Macro (loc, m, sl) ->
      let s =
        List.fold_right
          (fun (loc1, (_, s)) e ->
             let e1 = let loc = loc1 in <:expr< $str:s$ >> in
             let loc = encl_loc loc1 loc in
             <:expr< [$e1$ :: $e$] >>)
          sl <:expr< [] >>
      in
      (<:str_item< macro $str:m$ $s$ >>, loc)
  | Comment (loc, s) ->
      (<:str_item< comment $str:s$ >>, loc)
  | QuotedString (loc, s) ->
      (<:str_item< quoted_string $str:s$ >>, loc)
  | LeadingSpace (loc, s) ->
      (<:str_item< leading_space $str:s$ >>, loc)

let m4ltosil_default m4l = List.map m4tosi m4l
let m4ltosil = ref m4ltosil_default

let m4_item = Grammar.Entry.create gram "m4_item"
let argument = Grammar.Entry.create gram "argument"

EXTEND
  GLOBAL: implem m4_item argument;
  implem:
    [ [ m4l = m4_list -> (!m4ltosil m4l, false) ] ]
  ;
  m4_list:
    [ [ x = m4_item; l = SELF -> x :: l
      | EOI -> [] ] ]
  ;
  m4_item:
    [ [ (loc1, m) = name; "("; sl = LIST0 argument SEP ","; ")" ->
          Macro (loc1, m, sl)
      | (loc1, m) = name -> Macro (loc1, m, [])
      | s = COMMENT -> Comment (loc, s)
      | s = QUOTSTR -> QuotedString (loc, s)
      | s = LEADSP -> LeadingSpace (loc, s) ] ]
  ;
  name:
    [ [ m = NAME -> (loc, m) ] ]
  ;
  argument:
    [ [ s = LEADSP; a = arg -> (loc, (s, a))
      | a = arg -> (loc, ("", a)) ] ]
  ;
  arg:
    [ [ s = LEADSP; a = arg -> s ^ a
      | s = NAME; a = arg -> s ^ a
      | s = QUOTSTR; a = arg -> s ^ a
      | s = OTHER; a = arg -> s ^ a
      | "("; sl = LIST0 arg SEP ","; ")"; a = arg ->
         let s =
           List.fold_left (fun s a -> if s = "" then a else s ^ "," ^ a)
             "" sl
          in
          "(" ^ s ^ ")" ^ a
      | -> "" ] ]
  ;
END
