(* $Id: pa_syndex.ml 3445 2009-08-19 12:21:51Z stentzel $ *)

#load "q_MLast.cmo";;
#load "pa_macro.cmo";;

type mgc = Init | Loop | End | Undef

type m4 = Pa_m4.m4 =
    Macro of MLast.loc * string * (MLast.loc * (string * string)) list
  | Comment of MLast.loc * string
  | QuotedString of MLast.loc * string
  | LeadingSpace of MLast.loc * string

type info =
  { ctx : mgc }

let encl_loc loc1 loc2 =
  IFDEF CAMLP5 THEN Stdpp.encl_loc loc1 loc2
  ELSE (fst loc1, snd loc2)
  END

let loc_of_m4 =
  function
    Macro (loc, _, _) -> loc
  | Comment (loc, _) -> loc
  | QuotedString (loc, _) -> loc
  | LeadingSpace (loc, _) -> loc

let zero_of_type loc =
  function
    "bool" -> <:expr< False >>
  | "float" -> <:expr< 0.0 >>
  | "int" -> <:expr< 0 >>
  | t -> <:expr< $lid:"zero_of_" ^ t$ >>

let is_keyword =
  let ht = Hashtbl.create 503 in
  List.iter (fun s -> Hashtbl.add ht s s)
    ["endcase_"; "endif_"; "endloop_"; "endmain_"; "endswitch_";
     "endthread_"];
  fun s -> Hashtbl.mem ht s

let is_int s =
  try let (_ : int) = int_of_string s in true with
    Failure _ -> false

let is_float s =
  try let (_ : float) = float_of_string s in true with
    Failure _ -> false

let ensure_float_string v =
  if is_int v then string_of_float (float_of_string v) else v

let last_word s =
  try
    let i = String.rindex s '_' in
    String.sub s (i + 1) (String.length s - i - 1)
  with
    Not_found -> s

let shorten_names = ref false

let mod_name =
  let ht = Hashtbl.create 1009 in
  let hti = Hashtbl.create 1009 in
  fun s ->
    if !shorten_names then
      try Hashtbl.find ht s with
        Not_found ->
          let ini = last_word s in
          let rec loop cnt =
            let t = ini ^ "_" ^ string_of_int cnt in
            if Hashtbl.mem hti t then loop (cnt + 1) else t
          in
          let t = if Hashtbl.mem hti ini then loop 1 else ini in
          Hashtbl.add hti t s;
          Hashtbl.add ht s t;
          t
     else String.uncapitalize s

let fix_name s = String.uncapitalize s

let trd x = snd (snd x)

let make_str_list loc pl =
  List.fold_right (fun (loc, p) e -> <:expr< [$str:snd p$ :: $e$] >>) pl
    <:expr< [] >>

let make_int_list loc pl =
  List.fold_right (fun (loc, p) e -> <:expr< [$int:snd p$ :: $e$] >>) pl
    <:expr< [] >>

let make_flo_list loc pl =
  List.fold_right
    (fun (loc, p) e ->
       let p = ensure_float_string (snd p) in
       <:expr< [$flo:p$ :: $e$] >>)
    pl <:expr< [] >>

let make_int_var_list loc pl =
  List.fold_right
    (fun (i, v) e ->
       let loc = encl_loc (fst i) (fst v) in
       let i = let loc = fst i in <:expr< $int:trd i$ >> in
       let v = let loc = fst v in <:expr< $lid:mod_name (trd v)$ >> in
       <:expr< [($i$, $v$) :: $e$] >>)
    pl <:expr< [] >>

let delay_parameters loc pl =
  let (len, pl) =
    match pl with
      (_, s) :: pl -> (int_of_string (snd s), pl)
    | _ -> failwith "delay_parameters"
  in
  let (ini, pl) =
    let rec get_ini len pl =
      if len <= 0 then ([], pl)
      else
        match pl with
          p :: pl ->
            let (ini, pl) = get_ini (len - 1) pl in
            (p :: ini, pl)
        | [] -> failwith "delay_parameters"
    in
    get_ini len pl
  in
  let pl =
    match pl with
      (_, len2) :: pl ->
        if int_of_string (snd len2) <> len then failwith "delay_parameters";
        pl
    | [] -> failwith "delay_parameters"
  in
  let (newval, key) =
    match pl with
      [a; b] -> (a, b)
    | [a; _; b] -> (a, b)
    | _ -> failwith "delay_parameters"
  in
  (ini, newval, key)

let make_delay make_list info loc pl =
  let (ini, v, k) = delay_parameters loc pl in
  let k = let loc = fst k in <:expr< $lid:mod_name (trd k)$ >> in
  if info.ctx = Init then
    let ini = make_list loc ini in
    <:expr< ML_syndex.delay_init $ini$ $k$ >>
  else
    let v = let loc = fst v in <:expr< $lid:mod_name (trd v)$ >> in
    <:expr< ML_syndex.delay_shift $v$ $k$ >>

let make_condo loc pl =
  let (cond, pl) =
    match pl with
      cond :: pl -> (cond, pl)
    | [] -> Stdpp.raise_with_loc loc (Failure "make_condo")
  in
  let rec loop =
    function
      p :: e :: list ->
        let (list, key) = loop list in
        ((p, e) :: list, key)
    | [key] -> ([], key)
    | _ -> Stdpp.raise_with_loc loc (Failure "make_condo")
  in
  let (list, key) = loop pl in
  (cond, list, key)

let expr_of_expr_list loc =
  function
    [e] -> e
  | el -> <:expr< do { $list:el$ } >>

let rec expr info =
  parser
    [< 'Macro (loc, "CondO", pl) >] ->
      let (c, l, k) = make_condo loc pl in
      let cond = let loc = fst c in <:expr< $lid:mod_name (trd c)$ >> in
      let list = make_int_var_list loc l in
      let key = let loc = fst k in <:expr< $lid:mod_name (trd k)$ >> in
      <:expr< ML_syndex.condO $cond$ $list$ $key$ >>
  | [< 'Macro (loc, "if_", [k; v]); el = expr_list info;
       'Macro (loc, "endif_", []) ?? "endif_ expected" >] ->
      let k = let loc = fst k in <:expr< $lid:trd k$ >> in
      let v = let loc = fst v in <:expr< $int:trd v$ >> in
      let e = expr_of_expr_list loc el in
      <:expr< if $k$.(0) = $v$ then $e$ else () >>
  | [< 'Macro (loc, "loadDnto_", pl) >] ->
      let e = make_str_list loc pl in
      <:expr< ML_syndex.loadDnto $e$ >>
  | [< 'Macro (loc, "loadFrom_", pl) >] ->
      let e = make_str_list loc pl in
      <:expr< ML_syndex.loadFrom $e$ >>
  | [< 'Macro (loc, "Pre0_", p :: _) >] ->
      let s = let loc = fst p in <:expr< $str:mod_name (trd p)$ >> in
      let e = let loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_syndex.pre0 $s$ $e$ >>
  | [< 'Macro (loc, "Pre1_", p :: _) >] ->
      let s = let loc = fst p in <:expr< $str:mod_name (trd p)$ >> in
      let e = let loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_syndex.pre1 $s$ $e$ >>
  | [< 'Macro (loc, "recv_", bn :: st :: sn :: rnl) >] ->
      let bn = let loc = fst bn in <:expr< $lid:mod_name (trd bn)$ >> in
      let st = let loc = fst st in <:expr< $str:trd st$ >> in
      let sn = let loc = fst sn in <:expr< $str:trd sn$ >> in
      let rn = make_str_list loc rnl in
      <:expr< ML_syndex.recv $bn$ $st$ $sn$ $rn$ >>
  | [< 'Macro (loc, "recv_synchro_", _) >] ->
      <:expr< recv_synchro "..." >>
  | [< 'Macro (loc, "saveFrom_", pl) >] ->
      let e = make_str_list loc pl in
      <:expr< ML_syndex.saveFrom $e$ >>
  | [< 'Macro (loc, "saveUpto_", pl) >] ->
      let e = make_str_list loc pl in
      <:expr< ML_syndex.saveUpto $e$ >>
  | [< 'Macro (loc, "send_", bn :: st :: sn :: rnl) >] ->
      let bn = let loc = fst bn in <:expr< $lid:mod_name (trd bn)$ >> in
      let st = let loc = fst st in <:expr< $str:trd st$ >> in
      let sn = let loc = fst sn in <:expr< $str:trd sn$ >> in
      let rn = make_str_list loc rnl in
      <:expr< ML_syndex.send $bn$ $st$ $sn$ $rn$ >>
  | [< 'Macro (loc, "send_synchro_", _) >] ->
      <:expr< ML_syndex.send_synchro "..." >>
  | [< 'Macro (loc, "spawn_thread_", [p]) >] ->
      let e = let loc = fst p in <:expr< $lid:"com_thread_" ^ trd p$ >> in
      <:expr< ML_syndex.spawn_thread $e$ >>
  | [< 'Macro (loc, "Suc0_", p :: _) >] ->
      let s = let loc = fst p in <:expr< $str:mod_name (trd p)$ >> in
      let e = let loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_syndex.suc0 $s$ $e$ >>
  | [< 'Macro (loc, "Suc1_", p :: _) >] ->
      let s = let loc = fst p in <:expr< $str:mod_name (trd p)$ >> in
      let e = let loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_syndex.suc1 $s$ $e$ >>
  | [< 'Macro (loc1, "switch_", [(locs, s)]); pel = match_assoc_list info;
       'Macro (loc2, "endswitch_", []) ?? "endswitch_ expected" >] ->
      let e = let loc = locs in <:expr< $lid:mod_name (snd s)$ >> in
      let loc = encl_loc loc1 loc2 in
      let pel = pel @ [(<:patt< _ >>, <:vala< None >>, <:expr< () >>)] in
      <:expr< match $e$.(0) with [ $list:pel$ ] >>
  | [< 'Macro (loc, "sync_", _) >] ->
      <:expr< ML_syndex.sync "..." >>
  | [< 'Macro (loc, "wait_endthread_", [p]) >] ->
      let e = let loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_syndex.wait_endthread $e$ >>

  | [< 'Macro (loc, "float_delay", pl) >] ->
      make_delay make_flo_list info loc pl
  | [< 'Macro (loc, "int_delay", pl) >] ->
      make_delay make_int_list info loc pl

  | [< 'Macro (loc, "int_input", [i; p]) >] ->
      let n = let _loc = fst p in <:expr< $str:trd p$ >> in
      let i = let _loc = fst i in <:expr< $int:trd i$ >> in
      let p = let _loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_int.int_input $n$ $i$ $p$ >>
  | [< 'Macro (loc, "int_output", [i; p]) >] ->
      let n = let _loc = fst p in <:expr< $str:trd p$ >> in
      let i = let _loc = fst i in <:expr< $int:trd i$ >> in
      let p = let _loc = fst p in <:expr< $lid:mod_name (trd p)$ >> in
      <:expr< ML_int.int_output $n$ $i$ $p$ >>

  | [< 'Macro (loc, "operation_init_", (p :: pl)); s >] ->
      let loc = fst p in
      expr info [< 'Macro (loc, trd p, pl); s >]

  | [< 'Macro (loc, m, sl) when not (is_keyword m) >] ->
      let sl = if sl = [] then [(loc, ("", "0"))] else sl in
      List.fold_left
        (fun e (loc1, s) ->
           let f =
             let loc = loc1 in
             if is_int (snd s) then <:expr< $int:snd s$ >>
             else if is_float (snd s) then <:expr< $flo:snd s$ >>
             else <:expr< $lid:mod_name (snd s)$ >>
           in
           <:expr< $e$ $f$ >>)
        <:expr< $lid:fix_name m$ >> sl
and expr_list info =
  parser
    [< 'LeadingSpace (_, _); s >] -> expr_list info s
  | [< 'Macro (_, "int_input", _) when info.ctx <> Loop;
       el = expr_list info >] ->
      el
  | [< 'Macro (_, "int_output", _) when info.ctx <> Loop;
       el = expr_list info >] ->
      el
  | [< 'Macro (loc1, "loop_", []); el1 = expr_list {ctx = Loop};
       'Macro (loc2, "endloop_", []) ?? "endloop_ expected";
       el = expr_list {ctx = End} >] ->
      let loc = encl_loc loc1 loc2 in
      let e =
        <:expr<
           let rec loop n_iter_ =
             if ML_syndex.max_iter.val < 0 ||
                n_iter_ < ML_syndex.max_iter.val
             then do {
               $list:el1 @ [<:expr< loop (n_iter_ + 1) >>]$
             }
             else ()
           in loop 0 >>
      in
      if info.ctx = Init then
        (* inside main_ *)
        <:expr< ML_syndex.mgc.val := LOOP >> :: e ::
        <:expr< ML_syndex.mgc.val := END >> :: el
      else
        (* not inside main_, i.e. inside a com_thread *)
        e :: el
  | [< e = expr info; el = expr_list info >] -> e :: el
  | [< >] -> []
and match_assoc_list ctx =
  parser
    [< 'LeadingSpace (_, _); s >] -> match_assoc_list ctx s
  | [< pe = match_assoc ctx; pel = match_assoc_list ctx >] -> pe :: pel
  | [< >] -> []
and match_assoc ctx =
  parser
    [< 'Macro (loc1, "case_", [(locs, s)]); el = expr_list ctx;
       'Macro (loc2, "endcase_", []) ?? "endcase_ expected" >] ->
      let p = let loc = locs in <:patt< $int:snd s$ >> in
      let e = expr_of_expr_list (encl_loc loc1 loc2) el in
      (p, <:vala< None >>, e)

let struct_item =
  parser
    [< 'Macro (loc, "endprocessor_", _) >] ->
      let si = <:str_item< ML_syndex.endprocessor main >> in
      (si, loc)
  | [< 'Macro (loc, "include", [(_, (_, t))]) >] ->
      let t =
        if Filename.check_suffix t ".m4x" then
          Filename.chop_suffix t ".m4x"
        else t
      in
      (<:str_item< open $["ML_" ^ t]$ >>, loc)
  | [< 'Macro (loc, "sinclude", [(_, (_, t))]) >] ->
      let t =
        if Filename.check_suffix t ".m4x" then
          Filename.chop_suffix t ".m4x"
        else t
      in
      (<:str_item< open $["ML_" ^ t]$ >>, loc)
  | [< 'Macro (loc, "main_", []); el = expr_list {ctx = Init};
       'Macro (loc1, "endmain_", []) ?? "endmain_ expected" >] ->
      let e =
        expr_of_expr_list loc (<:expr< ML_syndex.mgc.val := INIT >> :: el)
      in
      (<:str_item< value main () = $e$ >>, loc)
  | [< 'Macro (loc, "processor_", [pt; pn; ap; tx; dt]) >] ->
      let fname = trd ap ^ ".cmi" in
      let mname = String.capitalize (trd ap) in
      let pt = let loc = fst pt in <:expr< $str:trd pt$ >> in
      let pn = let loc = fst pn in <:expr< $str:trd pn$ >> in
      let ap = let loc = fst ap in <:expr< $str:trd ap$ >> in
      let tx = let loc = fst tx in <:expr< $str:trd tx$ >> in
      let dt = let loc = fst dt in <:expr< $str:trd dt$ >> in
      let si = <:str_item< processor $pt$ $pn$ $ap$ $tx$ $dt$ >> in
      let si =
        let sil =
          if Sys.file_exists fname then
            [<:str_item< open $[mname]$ >>; si]
          else [si]
        in
        <:str_item< declare $list:sil$ end >>
      in
      (si, loc)
  | [< 'Macro (loc, "thread_", typ :: nam :: proc_list);
       el = expr_list {ctx = Undef};
       'Macro (loc1, "endthread_", []) ?? "endthread_ expected" >] ->
      let p = let loc = fst nam in <:patt< $lid:"com_thread_" ^ trd nam$ >> in
      let el = el @ [<:expr< Thread.exit () >>] in
      let e = expr_of_expr_list loc el in
      (<:str_item< value $p$ () = $e$ >>, loc)
  | [< e = expr {ctx = Undef} >] ->
      let loc = MLast.loc_of_expr e in
      (<:str_item< $exp:e$ >>, loc)

let rec struct_item_list =
  parser
    [< 'LeadingSpace (_, _); s >] -> struct_item_list s
  | [< 'Macro (loc, "alias_", [(locv1, v1); (locv2, v2); _; _]); strm >] ->
      let labn1 = mod_name (snd v1) in
      let labn2 = mod_name (snd v2) in
      let v1 = let loc = locv1 in <:patt< $lid:labn1$ >> in
      let v2 = let loc = locv2 in <:expr< $lid:labn2$ >> in
      let si = (<:str_item< value $v1$ = $v2$ >>, loc) in
      let sil = struct_item_list strm in
      si :: sil
  | [< 'Macro (loc, "alloc_", [(loct, typ); (locl, lab); siz]); strm >] ->
      let lab = let loc = locl in <:patt< $lid:mod_name (snd lab)$ >> in
      let zero = zero_of_type loct (snd typ) in
      let si =
        let e =
          let loc = fst siz in
          let len = int_of_string (trd siz) in
          if len = 1 then <:expr< [| $zero$ |] >>
          else <:expr< Array.create $int:trd siz$ $zero$ >>
        in
        (<:str_item< value $lab$ = $e$ >>, loc)
      in
      let sil = struct_item_list strm in
      si :: sil
  | [< 'Macro (loc, "semaphores_", pl); sil = struct_item_list >] ->
      let si =
        let (_, sil) =
          List.fold_left
            (fun (n, sil) (loc, (_, s)) ->
               if s = "" then (n, sil)
               else
                 let s = <:patt< $lid:mod_name s$ >> in
                 let sn = <:expr< $int:string_of_int n$ >> in
                 (n + 1, <:str_item< value $s$ = $sn$ >> :: sil))
            (0, []) pl
        in
        (<:str_item< declare $list:List.rev sil$ end >>, loc)
      in
      si :: sil
  | [< si = struct_item; sil = struct_item_list >] -> si :: sil
  | [< 'Macro (loc, m, _) >] ->
      Stdpp.raise_with_loc loc (Stream.Error ("bad macro: " ^ m))
  | [< 'm4 >] -> Stdpp.raise_with_loc (loc_of_m4 m4) (Stream.Error "bad macro")
  | [< >] -> []

let m4ltosil m4l =
  let strm = Stream.of_list m4l in
  try struct_item_list strm with
    exc ->
      let m4 =
        match strm with parser
          [< 'm4 >] -> m4
        | [< >] -> List.hd (List.rev m4l)
      in
      Stdpp.raise_with_loc (loc_of_m4 m4) exc

let _ = Pa_m4.m4ltosil := m4ltosil
let _ = Pcaml.add_option "-shnm" (Arg.Set shorten_names) "Shorten names"
